/**
 * Copyright (c) 2014 - 2017, Nordic Semiconductor ASA
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 * 
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 * 
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
/**
 * @brief BLE LED Button Service central and client application main file.
 *
 * This file contains the source code for a sample client application using the LED Button service.
 */

#include <app_button.h>
#include <app_error.h>
#include <app_timer.h>
#include <app_util.h>
#include <ble.h>
#include <bsp.h>
#include <bsp_btn_ble.h>
#include <ble_ias_c.h>
#include <ble_db_discovery.h>
#include <ble_err.h>
#include <ble_gap.h>
#include <ble_gattc.h>
#include <ble_gatts.h>
#include <ble_hci.h>
#include <ble_conn_state.h>
#include <ble_lbs_c.h>
#include <ble_stack_handler_types.h>
#include <boards.h>
#include <nrf_ble_gatt.h>
#include <nrf_error.h>
#include <nrf_sdm.h>
#include <nrf_soc.h>
#include <pca10056.h>
#include <sdk_errors.h>
#include <softdevice_handler.h>
#include <stdbool.h>
#include <string.h>
#include <sys/_stdint.h>

/* ADDED */

#include <ble_ias.h>
#include <fstorage.h>
#include <bsp.h>
#include <peer_manager.h>
#include <fds.h>

#define CONN_CFG_TAG                1                                   /**< A tag that refers to the BLE stack configuration we set with @ref sd_ble_cfg_set. Default tag is @ref BLE_CONN_CFG_TAG_DEFAULT. */

/* END ADDED */

#define NRF_LOG_MODULE_NAME "APP"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"

#define CENTRAL_LINK_COUNT        1                                     /**< Number of central links used by the application. When changing this number remember to adjust the RAM settings*/
#define PERIPHERAL_LINK_COUNT     0                                     /**< Number of peripheral links used by the application. When changing this number remember to adjust the RAM settings*/

#define CENTRAL_SCANNING_LED      BSP_BOARD_LED_0                       /**< Scanning LED will be on when the device is scanning. */
#define CENTRAL_CONNECTED_LED     BSP_BOARD_LED_1                       /**< Connected LED will be on when the device is connected. */

#define SCAN_INTERVAL             0x00A0                                /**< Determines scan interval in units of 0.625 millisecond. */
#define SCAN_WINDOW               0x0050                                /**< Determines scan window in units of 0.625 millisecond. */
#define SCAN_TIMEOUT              0x0000                                /**< Timout when scanning. 0x0000 disables timeout. */

#define MIN_CONNECTION_INTERVAL   MSEC_TO_UNITS(7.5, UNIT_1_25_MS)      /**< Determines minimum connection interval in milliseconds. */
#define MAX_CONNECTION_INTERVAL   MSEC_TO_UNITS(30, UNIT_1_25_MS)       /**< Determines maximum connection interval in milliseconds. */
#define SLAVE_LATENCY             0                                     /**< Determines slave latency in terms of connection events. */
#define SUPERVISION_TIMEOUT       MSEC_TO_UNITS(4000, UNIT_10_MS)       /**< Determines supervision time-out in units of 10 milliseconds. */

#define UUID16_SIZE               2                                     /**< Size of a UUID, in bytes. */

#define LEDBUTTON_LED             BSP_BOARD_LED_2                       /**< LED to indicate a change of state of the the Button characteristic on the peer. */
#define LEDBUTTON_BUTTON_PIN      BSP_BUTTON_0                          /**< Button that will write to the LED characteristic of the peer */
#define BUTTON_DETECTION_DELAY    APP_TIMER_TICKS(50)                   /**< Delay from a GPIOTE event until a button is reported as pushed (in number of timer ticks). */

#define NRF_LOG_USES_RTT    1

/* ADDED */

#define SEC_PARAM_BOND                  1                                       /**< Perform bonding. */
#define SEC_PARAM_MITM                  0                                       /**< Man In The Middle protection not required. */
#define SEC_PARAM_LESC                  0                                       /**< LE Secure Connections not enabled. */
#define SEC_PARAM_KEYPRESS              0                                       /**< Keypress notifications not enabled. */
#define SEC_PARAM_IO_CAPABILITIES       BLE_GAP_IO_CAPS_NONE                    /**< No I/O capabilities. */
#define SEC_PARAM_OOB                   0                                       /**< Out Of Band data not available. */
#define SEC_PARAM_MIN_KEY_SIZE          7                                       /**< Minimum encryption key size. */
#define SEC_PARAM_MAX_KEY_SIZE          16                                      /**< Maximum encryption key size. */

static ble_ias_t m_ias; /**< Structure used to identify the Immediate Alert service. */
static ble_ias_c_t m_ias_c; /**< Structure used to identify the client to the Immediate Alert Service at peer. */

static void on_ias_evt(ble_ias_t * p_ias, ble_ias_evt_t * p_evt);
static void on_ias_c_evt(ble_ias_c_t * p_lls, ble_ias_c_evt_t * p_evt);

static volatile bool m_is_ias_present = false; /**< Variable to indicate whether the immediate alert service has been discovered at the connected peer. */
static volatile bool m_is_high_alert_signalled; /**< Variable to indicate whether a high alert has been signalled to the peer. */

//static ble_gap_scan_params_t m_scan_param;                 /**< Scan parameters requested for scanning and connection. */
//static bool                  m_memory_access_in_progress;  /**< Flag to keep track of ongoing operations on persistent memory. */
//static bool                  m_whitelist_disabled;         /**< True if whitelist has been temporarily disabled. */

#define USING_RTT			1			/* 1 is true, 0 is false */


#include <SEGGER_RTT.h>



#define printf_RTT(...) \
do { \
     char str[64];\
     sprintf(str, __VA_ARGS__);\
     SEGGER_RTT_WriteString(0, str);\
 } while(0)


#if (USING_RTT == 1)
	#define printf printf_RTT
#else
	#define printf NRF_LOG_INFO
#endif


/* END ADDED */

/**@brief Variable length data encapsulation in terms of length and pointer to data.
 */
typedef struct {
	uint8_t * p_data; /**< Pointer to data. */
	uint16_t data_len; /**< Length of data. */
} data_t;

static char const m_target_periph_name[] = "ntkr"; /**< Name of the device we try to connect to. This name is searched in the scan report data*/

/**@brief Parameters used when scanning.
 */
static ble_gap_scan_params_t const m_scan_params = { .active = 1, .interval =
SCAN_INTERVAL, .window = SCAN_WINDOW, .timeout = SCAN_TIMEOUT,
#if (NRF_SD_BLE_API_VERSION <= 2)
		.selective = 0,
		.p_whitelist = NULL,
#endif
#if (NRF_SD_BLE_API_VERSION >= 3)
		.use_whitelist = 0,
#endif
		};


/**@brief Connection parameters requested for connection.
 */
static ble_gap_conn_params_t const m_connection_param = {
		(uint16_t) MIN_CONNECTION_INTERVAL, (uint16_t) MAX_CONNECTION_INTERVAL,
		(uint16_t) SLAVE_LATENCY, (uint16_t) SUPERVISION_TIMEOUT };

//static ble_lbs_c_t m_ble_lbs_c; /**< Main structure used by the LBS client module. */
static ble_db_discovery_t m_ble_db_discovery; /**< DB structure used by the database discovery module. */
static nrf_ble_gatt_t m_gatt; /**< GATT module instance. */

/**@brief Function to handle asserts in the SoftDevice.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num     Line number of the failing ASSERT call.
 * @param[in] p_file_name  File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name) {
	app_error_handler(0xDEADBEEF, line_num, p_file_name);
}


/**
 * @brief Parses advertisement data, providing length and location of the field in case
 *        matching data is found.
 *
 * @param[in]  type       Type of data to be looked for in advertisement data.
 * @param[in]  p_advdata  Advertisement report length and pointer to report.
 * @param[out] p_typedata If data type requested is found in the data report, type data length and
 *                        pointer to data will be populated here.
 *
 * @retval NRF_SUCCESS if the data type is found in the report.
 * @retval NRF_ERROR_NOT_FOUND if the data type could not be found.
 */
static uint32_t adv_report_parse(uint8_t type, data_t * p_advdata,
		data_t * p_typedata) {
	uint32_t index = 0;
	uint8_t * p_data;

	p_data = p_advdata->p_data;

	while (index < p_advdata->data_len) {
		uint8_t field_length = p_data[index];
		uint8_t field_type = p_data[index + 1];

		if (field_type == type) {
			p_typedata->p_data = &p_data[index + 2];
			p_typedata->data_len = field_length - 1;
			return NRF_SUCCESS;
		}
		index += field_length + 1;
	}
	return NRF_ERROR_NOT_FOUND;
}

/* ADDED */

// USEFUL

///**@brief Retrieve a list of peer manager peer IDs.
// *
// * @param[inout] p_peers   The buffer where to store the list of peer IDs.
// * @param[inout] p_size    In: The size of the @p p_peers buffer.
// *                         Out: The number of peers copied in the buffer.
// */
//static void peer_list_get(pm_peer_id_t * p_peers, uint32_t * p_size)
//{
//    pm_peer_id_t peer_id;
//    uint32_t     peers_to_copy;
//
//    peers_to_copy = (*p_size < BLE_GAP_WHITELIST_ADDR_MAX_COUNT) ?
//                     *p_size : BLE_GAP_WHITELIST_ADDR_MAX_COUNT;
//
//    peer_id = pm_next_peer_id_get(PM_PEER_ID_INVALID);
//    *p_size = 0;
//
//    while ((peer_id != PM_PEER_ID_INVALID) && (peers_to_copy--))
//    {
//        p_peers[(*p_size)++] = peer_id;
//        peer_id = pm_next_peer_id_get(peer_id);
//    }
//}


//static void whitelist_load()
//{
//    ret_code_t   ret;
//    pm_peer_id_t peers[8];
//    uint32_t     peer_cnt;
//
//    memset(peers, PM_PEER_ID_INVALID, sizeof(peers));
//    peer_cnt = (sizeof(peers) / sizeof(pm_peer_id_t));
//
//    // Load all peers from flash and whitelist them.
//    peer_list_get(peers, &peer_cnt);
//
//    ret = pm_whitelist_set(peers, peer_cnt);
//    APP_ERROR_CHECK(ret);
//
//    // Setup the device identies list.
//    // Some SoftDevices do not support this feature.
//    ret = pm_device_identities_list_set(peers, peer_cnt);
//    if (ret != NRF_ERROR_NOT_SUPPORTED)
//    {
//        APP_ERROR_CHECK(ret);
//    }
//}

/* END ADDED */


/**@brief Function to start scanning.
 */
static void scan_start(void) {

//    uint32_t flash_busy;
//
//	printf("b4 flash\n");
//
//	// TO BRING BACK, SEARCH FOR m_memory_access_in_progress
//
//    // If there is any pending write to flash, defer scanning until it completes.
//    (void) fs_queued_op_count_get(&flash_busy);
//
//    if (flash_busy != 0)
//    {
//    	printf("FLASH_BUSY != 0\n");
//        m_memory_access_in_progress = true;
//        return;
//    }
//
//	printf("after flash\n");
//
//    // Whitelist buffers.
//    ble_gap_addr_t whitelist_addrs[8];
//    ble_gap_irk_t  whitelist_irks[8];
//
//    memset(whitelist_addrs, 0x00, sizeof(whitelist_addrs));
//    memset(whitelist_irks,  0x00, sizeof(whitelist_irks));
//
//    uint32_t addr_cnt = (sizeof(whitelist_addrs) / sizeof(ble_gap_addr_t));
//    uint32_t irk_cnt  = (sizeof(whitelist_irks)  / sizeof(ble_gap_irk_t));
//
//	printf("1\n");
//
//    #if (NRF_SD_BLE_API_VERSION <= 2)
//
//        ble_gap_addr_t * p_whitelist_addrs[8];
//        ble_gap_irk_t  * p_whitelist_irks[8];
//
//        for (uint32_t i = 0; i < 8; i++)
//        {
//            p_whitelist_addrs[i] = &whitelist_addrs[i];
//            p_whitelist_irks[i]  = &whitelist_irks[i];
//        }
//
//        ble_gap_whitelist_t whitelist =
//        {
//            .pp_addrs = p_whitelist_addrs,
//            .pp_irks  = p_whitelist_irks,
//        };
//
//    #endif
//
//    // Reload the whitelist and whitelist all peers.
//    whitelist_load();
//
//	printf("2\n");
//
//
//    // Get the whitelist previously set using pm_whitelist_set().
//    err_code = pm_whitelist_get(whitelist_addrs, &addr_cnt,
//                           whitelist_irks,  &irk_cnt);
//
//    m_scan_param.active   = 0;
//    m_scan_param.interval = SCAN_INTERVAL;
//    m_scan_param.window   = SCAN_WINDOW;
//
//    if (((addr_cnt == 0) && (irk_cnt == 0)) ||
//        (m_whitelist_disabled))
//    {
//    	printf("Don't use whitelist.\n");
//        // Don't use whitelist.
//        #if (NRF_SD_BLE_API_VERSION <= 2)
//            m_scan_param.selective   = 0;
//            m_scan_param.p_whitelist = NULL;
//        #endif
//        #if (NRF_SD_BLE_API_VERSION >= 3)
//            m_scan_param.use_whitelist  = 0;
//            m_scan_param.adv_dir_report = 0;
//        #endif
//        m_scan_param.timeout  = 0x0000; // No timeout.
//    }
//    else
//    {
//    	printf("Use whitelist.\n");
//        // Use whitelist.
//        #if (NRF_SD_BLE_API_VERSION <= 2)
//            whitelist.addr_count     = addr_cnt;
//            whitelist.irk_count      = irk_cnt;
//            m_scan_param.selective   = 1;
//            m_scan_param.p_whitelist = &whitelist;
//        #endif
//        #if (NRF_SD_BLE_API_VERSION >= 3)
//            m_scan_param.use_whitelist  = 1;
//            m_scan_param.adv_dir_report = 0;
//        #endif
//        m_scan_param.timeout  = 0x001E; // 30 seconds.
//    }
//
//	printf("3\n");

	ret_code_t err_code;

	(void) sd_ble_gap_scan_stop();

	err_code = sd_ble_gap_scan_start(&m_scan_params);
	APP_ERROR_CHECK(err_code);


	bsp_indication_set(BSP_INDICATE_SCANNING);

}


/* ADDED */

/**@brief Function for handling Service errors.
 *
 * @details A pointer to this function will be passed to each service which may need to inform the
 *          application about an error.
 *
 * @param[in] nrf_error   Error code containing information about what went wrong.
 */
static void service_error_handler(uint32_t nrf_error) {
	APP_ERROR_HANDLER(nrf_error);
}

/**@brief Function for initializing the Immediate Alert Service.
 */
static void ias_init(void) {
	printf("in ias_init\n");
	ret_code_t err_code;
	ble_ias_init_t ias_init_obj;

	memset(&ias_init_obj, 0, sizeof(ias_init_obj));

	ias_init_obj.evt_handler = on_ias_evt;

	err_code = ble_ias_init(&m_ias, &ias_init_obj);

	printf("ias_init ble_ias_init err_code: %lu \r\n", err_code);
	APP_ERROR_CHECK(err_code);
}

/**@brief Function for initializing the immediate alert service client.
 *
 * @details This will initialize the client side functionality of the Find Me profile.
 */
static void ias_client_init(void) {
	printf("in ias_client_init\n");
	ret_code_t err_code;
	ble_ias_c_init_t ias_c_init_obj;

	memset(&ias_c_init_obj, 0, sizeof(ias_c_init_obj));

	m_is_high_alert_signalled = false;

	ias_c_init_obj.evt_handler = on_ias_c_evt;
	ias_c_init_obj.error_handler = service_error_handler;

	err_code = ble_ias_c_init(&m_ias_c, &ias_c_init_obj);
	APP_ERROR_CHECK(err_code);
}

/**@brief Function for initializing the services that will be used by the application.
 */


static void services_init(void) {

	// RINGING SERVICE SENDER?
	SEGGER_RTT_WriteString(0, "ias_init\n");
	ias_init();

	// SEPERATION ALERT SERVICE
	SEGGER_RTT_WriteString(0, "lls_init\n");
//    lls_init();

	// RINGING SERVICE RECEIVER?
	SEGGER_RTT_WriteString(0, "ias_client_init\n");
	ias_client_init();
}

// SET THE ALERT TYPE (LIGHT) (THIS IS THE NORDIC'S VERSION OF RINGING)

/**@brief Function for the Signals alert event from Immediate Alert or Link Loss services.
 *
 * @param[in] alert_level  Requested alert level.
 */
static void alert_signal(uint8_t alert_level) {
	ret_code_t err_code;

	switch (alert_level) {
	case BLE_CHAR_ALERT_LEVEL_NO_ALERT:
		NRF_LOG_INFO("No Alert.\r\n");
		err_code = bsp_indication_set(BSP_INDICATE_ALERT_OFF);
		APP_ERROR_CHECK(err_code);
		break; // BLE_CHAR_ALERT_LEVEL_NO_ALERT

	case BLE_CHAR_ALERT_LEVEL_MILD_ALERT:
		NRF_LOG_INFO("Mild Alert.\r\n");
		err_code = bsp_indication_set(BSP_INDICATE_ALERT_0);
		APP_ERROR_CHECK(err_code);
		break; // BLE_CHAR_ALERT_LEVEL_MILD_ALERT

	case BLE_CHAR_ALERT_LEVEL_HIGH_ALERT:
		NRF_LOG_INFO("HIGH Alert.\r\n");
		err_code = bsp_indication_set(BSP_INDICATE_ALERT_3);
		APP_ERROR_CHECK(err_code);
		break; // BLE_CHAR_ALERT_LEVEL_HIGH_ALERT

	default:
		// No implementation needed.
		break;
	}
}

// HANDLE INCOMING ALERT REQUESTS

/**@brief Function for handling Immediate Alert events.
 *
 * @details This function will be called for all Immediate Alert events which are passed to the
 *          application.
 *
 * @param[in] p_ias  Immediate Alert structure.
 * @param[in] p_evt  Event received from the Immediate Alert service.
 */
static void on_ias_evt(ble_ias_t * p_ias, ble_ias_evt_t * p_evt) {
	printf("*************************************************\n");
	printf("in on_ias_evt\n");
	switch (p_evt->evt_type) {
	case BLE_IAS_EVT_ALERT_LEVEL_UPDATED:
		printf("BLE_IAS_EVT_ALERT_LEVEL_UPDATED\n");
		alert_signal(p_evt->params.alert_level);
		break; // BLE_IAS_EVT_ALERT_LEVEL_UPDATED

	default:
		// No implementation needed.
		break;
	}
}

// HANDLE OUTGOING ALERT REQUESTS (REQUESTING THE CONNECTED DEVICE TO RING USING THE BUTTON EVENT HANDLER)

/**@brief Function for handling IAS Client events.
 *
 * @details This function will be called for all IAS Client events which are passed to the
 *          application.
 *
 * @param[in] p_ias_c  IAS Client structure.
 * @param[in] p_evt    Event received.
 */
static void on_ias_c_evt(ble_ias_c_t * p_ias_c, ble_ias_c_evt_t * p_evt) {
	ret_code_t err_code;

	printf("in on ias c evt\n");

	switch (p_evt->evt_type) {
	case BLE_IAS_C_EVT_DISCOVERY_COMPLETE:

		printf("in BLE_IAS_C_EVT_DISCOVERY_COMPLETE\n");
		// IAS is found on peer. The Find Me Locator functionality of this app will work.
		err_code = ble_ias_c_handles_assign(&m_ias_c, p_evt->conn_handle,
				p_evt->alert_level.handle_value);
		APP_ERROR_CHECK(err_code);

		printf("before setting m_is_ias_present to TRUE\n");
		m_is_ias_present = true;
		break; // BLE_IAS_C_EVT_DISCOVERY_COMPLETE

	case BLE_IAS_C_EVT_DISCOVERY_FAILED:
		printf("in BLE_IAS_C_EVT_DISCOVERY_FAILED\n");
		// IAS is not found on peer. The Find Me Locator functionality of this app will NOT work.
		break; // BLE_IAS_C_EVT_DISCOVERY_FAILED

	case BLE_IAS_C_EVT_DISCONN_COMPLETE:
		// Disable alert buttons
		m_is_ias_present = false;
		break; // BLE_IAS_C_EVT_DISCONN_COMPLETE

	default:
		break;
	}
}

/* END ADDED */

/**@brief Function for handling the advertising report BLE event.
 *
 * @param[in] p_ble_evt  Bluetooth stack event.
 */
static void on_adv_report(const ble_evt_t * const p_ble_evt) {
	ret_code_t err_code;
	data_t adv_data;
	data_t dev_name;
	bool do_connect = false;

	// For readibility.
	ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;
	ble_gap_addr_t const * peer_addr = &p_gap_evt->params.adv_report.peer_addr;

	// Initialize advertisement report for parsing
	adv_data.p_data = (uint8_t *) p_gap_evt->params.adv_report.data;
	adv_data.data_len = p_gap_evt->params.adv_report.dlen;

	// Search for advertising names.
	bool name_found = false;
	err_code = adv_report_parse(BLE_GAP_AD_TYPE_COMPLETE_LOCAL_NAME, &adv_data,
			&dev_name);


	if (err_code != NRF_SUCCESS) {
		// Look for the short local name if it was not found as complete.
		err_code = adv_report_parse(BLE_GAP_AD_TYPE_SHORT_LOCAL_NAME, &adv_data,
				&dev_name);
		if (err_code != NRF_SUCCESS) {
			// If we can't parse the data, then exit
			return;
		} else {
			name_found = true;
		}
	} else {
		name_found = true;
	}
//	printf("%d\n", *dev_name.p_data);

//	printf("name_found: %d\n", name_found);

	if (do_connect) printf("IT IS TRUE... WHYYYY");

	if (name_found) {
		if (strlen(m_target_periph_name) != 0) {
			if (memcmp(m_target_periph_name, dev_name.p_data, dev_name.data_len)
					== 0) {
				printf("FOUND THE WANTED DEVICE NAME\n");
				do_connect = true;
			}
		}
	}

	if (do_connect) {
		// Initiate connection.

//		#if (NRF_SD_BLE_API_VERSION <= 2)
//			m_scan_param.selective = 0;
//		#endif
//		#if (NRF_SD_BLE_API_VERSION >= 3)
//			m_scan_param.use_whitelist = 0;
//		#endif

		printf("in do_connect\n");
		err_code = sd_ble_gap_connect(peer_addr, &m_scan_params,
				&m_connection_param,
				BLE_CONN_CFG_TAG_DEFAULT);

//        m_whitelist_disabled = false;

        if (err_code != NRF_SUCCESS)
        {
            NRF_LOG_ERROR("Connection Request Failed, reason %d.\r\n", err_code);
        }

//		printf("sd ble gap connect err_code: %lu \r\n", err_code);
		APP_ERROR_CHECK(err_code);
	}
}

/**@brief Function for handling BLE Stack events concerning central applications.
 *
 * @details This function keeps the connection handles of central applications up-to-date. It
 * parses scanning reports, initiating a connection attempt to peripherals when a target UUID
 * is found, and manages connection parameter update requests. Additionally, it updates the status
 * of LEDs used to report central applications activity.
 *
 * @note Since this function updates connection handles, @ref BLE_GAP_EVT_DISCONNECTED events
 *       should be dispatched to the target application before invoking this function.
 *
 * @param[in] p_ble_evt Bluetooth stack event.
 */
static void on_ble_evt(ble_evt_t const * p_ble_evt) {

	printf("on_ble_evt\n");

	ret_code_t err_code;

	// For readability.
	ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;

	switch (p_ble_evt->header.evt_id) {

	case BLE_GAP_EVT_CONNECTED: {

        printf("Connected. on_ble_evt\r\n");

//        err_code = ble_lbs_c_handles_assign(&m_ble_lbs_c, p_gap_evt->conn_handle, NULL);
//        APP_ERROR_CHECK(err_code);
//
//        printf("ble_lbs_c_handles_assign err_code: %lu \r\n", err_code);


		memset(&m_ble_db_discovery, 0x00, sizeof(m_ble_db_discovery));
		err_code = ble_db_discovery_start(&m_ble_db_discovery,
				p_ble_evt->evt.gap_evt.conn_handle);

		// Discover peer's services.
//        err_code = ble_db_discovery_start(&m_ble_db_discovery, p_gap_evt->conn_handle);
//        APP_ERROR_CHECK(err_code);

		printf("ble_db_discovery_start err_code: %lu \r\n", err_code);
		APP_ERROR_CHECK(err_code);

        // Update LEDs status, and check if we should be looking for more
        // peripherals to connect to.
//        bsp_board_led_on(CENTRAL_CONNECTED_LED);
//        bsp_board_led_off(CENTRAL_SCANNING_LED);

    	bsp_indication_set(BSP_INDICATE_CONNECTED);

	}
		break; // BLE_GAP_EVT_CONNECTED

	case BLE_GAP_EVT_DISCONNECTED:
		NRF_LOG_INFO("Disconnected. on_ble_evt\r\n");
		err_code = bsp_indication_set(BSP_INDICATE_IDLE);
		APP_ERROR_CHECK(err_code);
		break; // BLE_GAP_EVT_DISCONNECTED

	case BLE_GAP_EVT_ADV_REPORT: {
		on_adv_report(p_ble_evt);
	}
		break;

	case BLE_GAP_EVT_TIMEOUT: {
        printf("BLE_GAP_EVT_TIMEOUT. on_ble_evt\r\n");
		// We have not specified a timeout for scanning, so only connection attemps can timeout.
		if (p_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN) {
			NRF_LOG_DEBUG("Connection request timed out.\r\n");
		}
	}
		break;

	case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST: {
        printf("BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST. on_ble_evt\r\n");
		// Accept parameters requested by peer.
//        printf(" ", p_gap_evt->params.conn_param_update_request.conn_params);
		err_code = sd_ble_gap_conn_param_update(p_gap_evt->conn_handle,
				&p_gap_evt->params.conn_param_update_request.conn_params);
		printf("sd_ble_gap_conn_param_update err_code: %lu \r\n", err_code);
		APP_ERROR_CHECK(err_code);
	}
		break;

	case BLE_GATTC_EVT_TIMEOUT: {
        printf("BLE_GATTC_EVT_TIMEOUT. on_ble_evt\r\n");
		// Disconnect on GATT Client timeout event.
		NRF_LOG_DEBUG("GATT Client Timeout.\r\n");
		err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
		BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
		APP_ERROR_CHECK(err_code);
	}
		break;

	case BLE_GATTS_EVT_TIMEOUT: {
        printf("BLE_GATTS_EVT_TIMEOUT. on_ble_evt\r\n");
		// Disconnect on GATT Server timeout event.
		NRF_LOG_DEBUG("GATT Server Timeout.\r\n");
		err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
		BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
		APP_ERROR_CHECK(err_code);
	}
		break;

	default:
		// No implementation needed.
		break;
	}
}


/* ADDED */

/**@brief Function for handling the Application's system events.
 *
 * @param[in]   sys_evt   system event.
 */
static void on_sys_evt(uint32_t sys_evt)
{
    switch (sys_evt)
    {
        case NRF_EVT_FLASH_OPERATION_SUCCESS:
            /* fall through */
        case NRF_EVT_FLASH_OPERATION_ERROR:

//            if (m_memory_access_in_progress)
//            {
//                m_memory_access_in_progress = false;
//                scan_start();
//            }
            break;

        default:
            // No implementation needed.
            break;
    }
}

/* END ADDED */


/**@brief Function for dispatching a BLE stack event to all modules with a BLE stack event handler.
 *
 * @details This function is called from the scheduler in the main loop after a BLE stack event has
 * been received.
 *
 * @param[in] p_ble_evt Bluetooth stack event.
 */
static void ble_evt_dispatch(ble_evt_t * p_ble_evt) {
	printf("ble_evt_dispatch\n");

	ble_conn_state_on_ble_evt(p_ble_evt);
    pm_on_ble_evt(p_ble_evt);
    ble_ias_on_ble_evt(&m_ias, p_ble_evt);
    ble_ias_c_on_ble_evt(&m_ias_c, p_ble_evt);
//    ble_lls_on_ble_evt(&m_lls, p_ble_evt);
    bsp_btn_ble_on_ble_evt(p_ble_evt);
	on_ble_evt(p_ble_evt);
	ble_db_discovery_on_ble_evt(&m_ble_db_discovery, p_ble_evt);
//	ble_lbs_c_on_ble_evt(&m_ble_lbs_c, p_ble_evt);
	nrf_ble_gatt_on_ble_evt(&m_gatt, p_ble_evt);

    // NEED?
//    bsp_btn_ble_on_ble_evt(p_ble_evt);

}


/* ADDED */

/**@brief Function for dispatching a system event to interested modules.
 *
 * @details This function is called from the System event interrupt handler after a system
 *          event has been received.
 *
 * @param[in]   sys_evt   System stack event.
 */
static void sys_evt_dispatch(uint32_t sys_evt)
{
	// Forward Softdevice events to the fstorage module
    fs_sys_event_handler(sys_evt);
    on_sys_evt(sys_evt);
}

/* END ADDED */

/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupts.
 */
static void ble_stack_init(void) {
	ret_code_t err_code;

	nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;

	// Initialize the SoftDevice handler module.
	SOFTDEVICE_HANDLER_INIT(&clock_lf_cfg, NULL);

	// Fetch the start address of the application RAM.
	uint32_t ram_start = 0;
	err_code = softdevice_app_ram_start_get(&ram_start);
	APP_ERROR_CHECK(err_code);
//	0x%x
	printf("RAM_START BLE_STACK_INIT: %lx\r\n", ram_start);

	// Overwrite some of the default configurations for the BLE stack.
	ble_cfg_t ble_cfg;

	/* MIGHT NEED TO INCLUDE */
	// Configure the number of custom UUIDS.
	memset(&ble_cfg, 0, sizeof(ble_cfg));
	// MAYBE CHANGE
	ble_cfg.common_cfg.vs_uuid_cfg.vs_uuid_count = 1;
	ble_cfg.common_cfg.vs_uuid_cfg.vs_uuid_count = 1;
	err_code = sd_ble_cfg_set(BLE_COMMON_CFG_VS_UUID, &ble_cfg, ram_start);
	APP_ERROR_CHECK(err_code);

	// Configure the maximum number of connections.
	memset(&ble_cfg, 0, sizeof(ble_cfg));
	ble_cfg.gap_cfg.role_count_cfg.periph_role_count = 1;
	ble_cfg.gap_cfg.role_count_cfg.central_role_count = CENTRAL_LINK_COUNT;
	ble_cfg.gap_cfg.role_count_cfg.central_sec_count = 1;
	err_code = sd_ble_cfg_set(BLE_GAP_CFG_ROLE_COUNT, &ble_cfg, ram_start);
	APP_ERROR_CHECK(err_code);

// 			 CHANGED TO:
//	memset(&ble_cfg, 0, sizeof(ble_cfg));
//	ble_cfg.gap_cfg.role_count_cfg.periph_role_count = 0;
//	ble_cfg.gap_cfg.role_count_cfg.central_role_count = CENTRAL_LINK_COUNT;
//	ble_cfg.gap_cfg.role_count_cfg.central_sec_count = 1;
//	err_code = sd_ble_cfg_set(BLE_GAP_CFG_ROLE_COUNT, &ble_cfg, ram_start);
//	APP_ERROR_CHECK(err_code);




	// Configure the maximum ATT MTU.
	memset(&ble_cfg, 0x00, sizeof(ble_cfg));
	ble_cfg.conn_cfg.conn_cfg_tag = CONN_CFG_TAG;
	ble_cfg.conn_cfg.params.gatt_conn_cfg.att_mtu = NRF_BLE_GATT_MAX_MTU_SIZE;
	err_code = sd_ble_cfg_set(BLE_CONN_CFG_GATT, &ble_cfg, ram_start);
	APP_ERROR_CHECK(err_code);

	// Change GATTS table size
	memset(&ble_cfg, 0, sizeof(ble_cfg));
	ble_cfg.gatts_cfg.attr_tab_size.attr_tab_size =
			BLE_GATTS_ATTR_TAB_SIZE_DEFAULT;
//	ble_cfg.gatts_cfg.attr_tab_size.attr_tab_size = BLE_GATTS_ATTR_TAB_SIZE_MIN;
	err_code = sd_ble_cfg_set(BLE_GATTS_CFG_ATTR_TAB_SIZE, &ble_cfg, ram_start);
	APP_ERROR_CHECK(err_code);

	// Enable BLE stack.
	err_code = softdevice_enable(&ram_start);
	APP_ERROR_CHECK(err_code);

	// Register with the SoftDevice handler module for BLE events.
	err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
	APP_ERROR_CHECK(err_code);

	/* ADDED */

    // Register with the SoftDevice handler module for System events.
    err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);

    /* END ADDED */

}

/**@brief Function for handling Peer Manager events.
 *
 * @param[in] p_evt  Peer Manager event.
 */
static void pm_evt_handler(pm_evt_t const * p_evt) {

	printf("pm_evt_handler ************************************************ \n");

	ret_code_t err_code;

	switch (p_evt->evt_id) {
	case PM_EVT_BONDED_PEER_CONNECTED: {
		NRF_LOG_INFO("Connected to a previously bonded device.\r\n");
		printf("Connected to a previously bonded device.\r\n");
	}
		break;

	case PM_EVT_CONN_SEC_SUCCEEDED: {
		NRF_LOG_INFO(
				"Connection secured: role: %d, conn_handle: 0x%x, procedure: %d.\r\n",
				ble_conn_state_role(p_evt->conn_handle), p_evt->conn_handle,
				p_evt->params.conn_sec_succeeded.procedure);
		printf(
				"Connection secured: role: %d, conn_handle: 0x%x, procedure: %d.\r\n",
				ble_conn_state_role(p_evt->conn_handle), p_evt->conn_handle,
				p_evt->params.conn_sec_succeeded.procedure);
	}
		break;

	case PM_EVT_CONN_SEC_FAILED: {
		/* Often, when securing fails, it shouldn't be restarted, for security reasons.
		 * Other times, it can be restarted directly.
		 * Sometimes it can be restarted, but only after changing some Security Parameters.
		 * Sometimes, it cannot be restarted until the link is disconnected and reconnected.
		 * Sometimes it is impossible, to secure the link, or the peer device does not support it.
		 * How to handle this error is highly application dependent. */
	}
		break;

	case PM_EVT_CONN_SEC_CONFIG_REQ: {
		// Reject pairing request from an already bonded peer.
		pm_conn_sec_config_t conn_sec_config = { .allow_repairing = false };
		pm_conn_sec_config_reply(p_evt->conn_handle, &conn_sec_config);
	}
		break;

	case PM_EVT_STORAGE_FULL: {
		// Run garbage collection on the flash.
		err_code = fds_gc();
		if (err_code == FDS_ERR_BUSY
				|| err_code == FDS_ERR_NO_SPACE_IN_QUEUES) {
			// Retry.
		} else {
			APP_ERROR_CHECK(err_code);
		}
	}
		break;

	case PM_EVT_PEERS_DELETE_SUCCEEDED: {
        scan_start();
	}
		break;

	case PM_EVT_LOCAL_DB_CACHE_APPLY_FAILED: {
		// The local database has likely changed, send service changed indications.
		pm_local_database_has_changed();
	}
		break;

	case PM_EVT_PEER_DATA_UPDATE_FAILED: {
		// Assert.
		APP_ERROR_CHECK(p_evt->params.peer_data_update_failed.error);
	}
		break;

	case PM_EVT_PEER_DELETE_FAILED: {
		// Assert.
		APP_ERROR_CHECK(p_evt->params.peer_delete_failed.error);
	}
		break;

	case PM_EVT_PEERS_DELETE_FAILED: {
		// Assert.
		APP_ERROR_CHECK(p_evt->params.peers_delete_failed_evt.error);
	}
		break;

	case PM_EVT_ERROR_UNEXPECTED: {
		// Assert.
		APP_ERROR_CHECK(p_evt->params.error_unexpected.error);
	}
		break;

	case PM_EVT_CONN_SEC_START:
	case PM_EVT_PEER_DATA_UPDATE_SUCCEEDED:
	case PM_EVT_PEER_DELETE_SUCCEEDED:
	case PM_EVT_LOCAL_DB_CACHE_APPLIED:
	case PM_EVT_SERVICE_CHANGED_IND_SENT:
	case PM_EVT_SERVICE_CHANGED_IND_CONFIRMED:
	default:
		break;
	}
}

/**@brief Function for the Peer Manager initialization.
 */
static void peer_manager_init(void) {
	ble_gap_sec_params_t sec_param;
	ret_code_t err_code;

	err_code = pm_init();
	APP_ERROR_CHECK(err_code);

	memset(&sec_param, 0, sizeof(ble_gap_sec_params_t));

	// Security parameters to be used for all security procedures.
	sec_param.bond = SEC_PARAM_BOND;
	sec_param.mitm = SEC_PARAM_MITM;
	sec_param.lesc = SEC_PARAM_LESC;
	sec_param.keypress = SEC_PARAM_KEYPRESS;
	sec_param.io_caps = SEC_PARAM_IO_CAPABILITIES;
	sec_param.oob = SEC_PARAM_OOB;
	sec_param.min_key_size = SEC_PARAM_MIN_KEY_SIZE;
	sec_param.max_key_size = SEC_PARAM_MAX_KEY_SIZE;
	sec_param.kdist_own.enc = 1;
	sec_param.kdist_own.id = 1;
	sec_param.kdist_peer.enc = 1;
	sec_param.kdist_peer.id = 1;

	err_code = pm_sec_params_set(&sec_param);
	APP_ERROR_CHECK(err_code);

	err_code = pm_register(pm_evt_handler);
	APP_ERROR_CHECK(err_code);
}


/* END ADDED */


/** @brief Clear bonding information from persistent storage
 */
static void delete_bonds(void)
{
    ret_code_t err_code;

    NRF_LOG_INFO("Erase bonds!\r\n");

    err_code = pm_peers_delete();
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for putting the chip into sleep mode.
 *
 * @note This function will not return.
 */
static void sleep_mode_enter(void)
{
    ret_code_t err_code;

    err_code = bsp_indication_set(BSP_INDICATE_IDLE);
    APP_ERROR_CHECK(err_code);

    // Prepare wakeup buttons.
    err_code = bsp_btn_ble_sleep_mode_prepare();
    APP_ERROR_CHECK(err_code);

    // Go to system-off mode (this function will not return; wakeup will cause a reset).
    err_code = sd_power_system_off();
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for handling events from the BSP module.
 *
 * @param[in]   event   Event generated when button is pressed.
 */
static void bsp_event_handler(bsp_event_t event)
{
    ret_code_t err_code;

    switch (event)
    {
        case BSP_EVENT_SLEEP:
            sleep_mode_enter();
            break;

        case BSP_EVENT_DISCONNECT:
            err_code = sd_ble_gap_disconnect(m_ias_c.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            if (err_code != NRF_ERROR_INVALID_STATE)
            {
                APP_ERROR_CHECK(err_code);
            }
            break;

        case BSP_EVENT_KEY_1:
        {
//          NRF_LOG_INFO("Ringing device\r\n");
            printf("Ringing device\r\n");
            if (m_is_ias_present)
            {
            	printf("in m_is_ias_present\n");
                if (!m_is_high_alert_signalled)
                {
                    printf("in !m_is_high_alert_signalled \r\n");
                    err_code =
                        ble_ias_c_send_alert_level(&m_ias_c, BLE_CHAR_ALERT_LEVEL_HIGH_ALERT);
                }
                else
                {
                    printf("in ELSE !m_is_high_alert_signalled \r\n");
                    err_code = ble_ias_c_send_alert_level(&m_ias_c, BLE_CHAR_ALERT_LEVEL_NO_ALERT);
                }

                if (err_code == NRF_SUCCESS)
                {
                    printf("in err_code == NRF_SUCCESS \r\n");
                    m_is_high_alert_signalled = !m_is_high_alert_signalled;
                }
                else if (
                    (err_code != NRF_ERROR_RESOURCES)
                    &&
                    (err_code != BLE_ERROR_GATTS_SYS_ATTR_MISSING)
                    &&
                    (err_code != NRF_ERROR_NOT_FOUND)
                        )
                {
                	printf("ble_ias_c_send_alert_level err_code: %lu \r\n", err_code);
                    APP_ERROR_HANDLER(err_code);
                }
            }
        } break;

        case BSP_EVENT_KEY_2:
            NRF_LOG_INFO("Alert Off.\r\n");
            err_code = bsp_indication_set(BSP_INDICATE_ALERT_OFF);
            APP_ERROR_CHECK(err_code);
            break;

        default:
            break;
    }
}

/**@brief Function for initializing buttons and leds.
 *
 * @param[out] p_erase_bonds  Will be true if the clear bonding button was pressed to wake the application up.
 */
static void buttons_leds_init(bool * p_erase_bonds)
{
    ret_code_t err_code;
    bsp_event_t startup_event;

    err_code = bsp_init(BSP_INIT_LED | BSP_INIT_BUTTONS, bsp_event_handler);
    APP_ERROR_CHECK(err_code);

    err_code = bsp_btn_ble_init(NULL, &startup_event);
    APP_ERROR_CHECK(err_code);

    *p_erase_bonds = (startup_event == BSP_EVENT_CLEAR_BONDING_DATA);

}
/* END ADDED */

static void db_disc_handler(ble_db_discovery_evt_t * p_evt) {
    ble_ias_c_on_db_disc_evt(&m_ias_c, p_evt);
//	ble_lbs_on_db_disc_evt(&m_ble_lbs_c, p_evt);
}

/**@brief Database discovery initialization.
 */
static void db_discovery_init(void) {
	ret_code_t err_code = ble_db_discovery_init(db_disc_handler);
	APP_ERROR_CHECK(err_code);
}

/**@brief Function to sleep until a BLE event is received by the application.
 */
static void power_manage(void) {
	ret_code_t err_code = sd_app_evt_wait();
	APP_ERROR_CHECK(err_code);
}

/**@brief Function for initializing the log.
 */
static void log_init(void) {
	ret_code_t err_code = NRF_LOG_INIT(NULL);
	APP_ERROR_CHECK(err_code);
}

/**@brief Function for initializing the timer.
 */
static void timer_init(void) {
	ret_code_t err_code = app_timer_init();
	APP_ERROR_CHECK(err_code);
}

/**@brief Function for initializing the GATT module.
 */
static void gatt_init(void) {
	ret_code_t err_code = nrf_ble_gatt_init(&m_gatt, NULL);
	APP_ERROR_CHECK(err_code);
}

int main(void) {

	/* ADDED */
    bool erase_bonds;
	/* END ADDED */

// Need this to happen before any NRF_LOG_INFO calls
	log_init();

	printf("TrackR Phone started.\r\n");

	timer_init();

	/* ADDED */
    buttons_leds_init(&erase_bonds);
	/* END ADDED */

	/* REMOVE POSSIBLY */
//	leds_init();
//	buttons_init();
	/* END REMOVE */

	ble_stack_init();

	gatt_init();
	printf("gatt done\n");

	/* ADDED */
	peer_manager_init();
	printf("peer manager done\r\n");

	db_discovery_init();
	printf("db discovery done\r\n");

	/* ADDED */
	services_init();
	printf("services done\r\n");


	/* END ADDED */

	/* REMOVE POSSIBLY */
//	lbs_c_init();
	/* END REMOVE */

//	SEGGER_RTT_WriteString(0, "lbs c done done\r\n");

	// Start scanning for peripherals and initiate connection to devices which
	// advertise the target name.
    if (erase_bonds == true)
    {
    	printf("DELETING BONDS");
        delete_bonds();
        // Scanning is started by PM_EVT_PEERS_DELETE_SUCCEEDED
    }
    else
    {
        scan_start();
    }


	// Turn on the LED to signal scanning.
//	bsp_board_led_on(CENTRAL_SCANNING_LED);


    printf("starting main loop\r\n");
	for (;;) {
		if (NRF_LOG_PROCESS() == false) {
			// Wait for BLE events.
			power_manage();
		}
	}
}